@@ -107,7 +107,7 @@
 #  define KEY_NAME_PTR(key) (key)->name
 #  define KEY_NAME_LENGTH(key) strlen((key)->name)
 #  define KEY_NAME_EQUAL_KEY(key1, key2)        \
-  strcmp((key1)->name, (key2)->name) == 0
+  (strcmp((key1)->name, (key2)->name) == 0)
 #  define KEY_NAME_FORMAT "%s"
 #  define KEY_NAME_FORMAT_VALUE(key) (key)->name
 #endif
@@ -215,12 +215,12 @@
 #elif defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 100000
 #  define mrn_init_alloc_root(root, name, block_size, pre_alloc_size, flags) \
   init_alloc_root(root, block_size, pre_alloc_size, flags)
-#elif MYSQL_VERSION_ID >= 50706
+#elif MYSQL_VERSION_ID >= 80027
 #  define mrn_init_alloc_root(root, name, block_size, pre_alloc_size, flags) \
-  init_alloc_root(mrn_memory_key, root, block_size, pre_alloc_size)
+  ::new ((void *)root) MEM_ROOT(mrn_memory_key, block_size)
 #else
 #  define mrn_init_alloc_root(root, name, block_size, pre_alloc_size, flags) \
-  init_alloc_root(root, block_size, pre_alloc_size)
+  init_alloc_root(mrn_memory_key, root, block_size, pre_alloc_size)
 #endif
 
 #if MYSQL_VERSION_ID < 100002 || !defined(MRN_MARIADB_P)
@@ -536,18 +536,8 @@ typedef uint mrn_srid;
 #  endif
 #endif
 
-#ifdef MRN_MARIADB_P
-#  define MRN_ABORT_ON_WARNING(thd) thd->abort_on_warning
-#else
-#  if MYSQL_VERSION_ID >= 50706
-#    define MRN_ABORT_ON_WARNING(thd) thd->is_strict_mode()
-#  else
-#    define MRN_ABORT_ON_WARNING(thd) thd->abort_on_warning
-#  endif
-#endif
-
 #define MRN_ERROR_CODE_DATA_TRUNCATE(thd)                               \
-  (MRN_ABORT_ON_WARNING(thd) ? ER_WARN_DATA_OUT_OF_RANGE : WARN_DATA_TRUNCATED)
+  (thd->is_strict_mode() ? ER_WARN_DATA_OUT_OF_RANGE : WARN_DATA_TRUNCATED)
 
 #if defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 100000
 #  define mrn_strconvert(from_cs,               \
@@ -997,12 +987,21 @@ typedef uint mrn_srid;
   (memset((table_share), 0, sizeof(TABLE_SHARE)))
 #endif
 
-#if MYSQL_VERSION_ID >= 80019 && !defined(MRN_MARIADB_P)
-#  define MRN_BITMAP_INIT(map, buf, n_bits) \
+#if !defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 80019
+#  define mrn_bitmap_init(map, buf, n_bits) \
     bitmap_init((map), (buf), (n_bits))
+#  define mrn_bitmap_free(map) \
+    bitmap_free((map))
+#elif defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 100700
+#  define mrn_bitmap_init(map, buf, n_bits) \
+    my_bitmap_init((map), (buf), (n_bits))
+#  define mrn_bitmap_free(map) \
+    my_bitmap_free((map))
 #else
-#  define MRN_BITMAP_INIT(map, buf, n_bits) \
+#  define mrn_bitmap_init(map, buf, n_bits) \
     bitmap_init((map), (buf), (n_bits), false)
+#  define mrn_bitmap_free(map) \
+    bitmap_free((map))
 #endif
 
 #if defined(MRN_MARIADB_P) && ((MYSQL_VERSION_ID >= 100237 && MYSQL_VERSION_ID < 100300) || \
@@ -1031,3 +1030,15 @@ typedef uint mrn_srid;
 #  define MRN_MATCH_ITEM_FLAGS(match_item) \
   ((match_item)->flags)
 #endif
+
+#if MYSQL_VERSION_ID >= 80027 && !defined(MRN_MARIADB_P)
+#  define MRN_FREE_ROOT(root) (root)->Clear()
+#else
+#  define MRN_FREE_ROOT(root) free_root((root), MYF(0))
+#endif
+
+#if (MYSQL_VERSION_ID >= 80000 && !defined(MRN_MARIADB_P))
+#  define MRN_LIST_SIZE(list) ((list)->size())
+#else
+#  define MRN_LIST_SIZE(list) ((list)->elements)
+#endif
